import logging
import sys

import pytest
from playwright.async_api import Error as PlaywrightError
from scrapy import Spider
from scrapy.http.headers import Headers
from scrapy_playwright._utils import _get_page_content, _NAVIGATION_ERROR_MSG, _encode_body


# page content retrieval
# ======================


@pytest.mark.skipif(sys.version_info < (3, 8), reason="AsyncMock was added on Python 3.8")
@pytest.mark.asyncio
async def test_get_page_content_ok():
    from unittest.mock import AsyncMock

    expected_content = "lorem ipsum"
    page = AsyncMock()
    page.content.return_value = expected_content
    content = await _get_page_content(
        page=page,
        spider=Spider("foo"),
        context_name="context",
        scrapy_request_url="https://example.org",
        scrapy_request_method="GET",
    )
    assert content == expected_content


@pytest.mark.skipif(sys.version_info < (3, 8), reason="AsyncMock was added on Python 3.8")
@pytest.mark.asyncio
async def test_get_page_content_retry_known_exception(caplog):
    from unittest.mock import AsyncMock

    caplog.set_level(logging.DEBUG)
    expected_content = "lorem ipsum"
    page = AsyncMock()
    page.url = "FAKE URL"
    page.content.side_effect = [PlaywrightError(_NAVIGATION_ERROR_MSG), expected_content]
    content = await _get_page_content(
        page=page,
        spider=Spider("foo"),
        context_name="context",
        scrapy_request_url="https://example.org",
        scrapy_request_method="GET",
    )
    assert content == expected_content
    assert (
        "scrapy-playwright",
        logging.DEBUG,
        f"Retrying to get content from page '{page.url}', error: 'Unable to retrieve"
        " content because the page is navigating and changing the content.'",
    ) in caplog.record_tuples


@pytest.mark.skipif(sys.version_info < (3, 8), reason="AsyncMock was added on Python 3.8")
@pytest.mark.asyncio
async def test_get_page_content_reraise_unknown_exception():
    from unittest.mock import AsyncMock

    expected_exception_message = "nope"
    page = AsyncMock()
    page.content.side_effect = PlaywrightError(expected_exception_message)
    with pytest.raises(PlaywrightError, match=expected_exception_message):
        await _get_page_content(
            page=page,
            spider=Spider("foo"),
            context_name="context",
            scrapy_request_url="https://example.org",
            scrapy_request_method="GET",
        )


# body encoding
# =============


def body_str(charset: str, content: str = "áéíóú") -> str:
    return f"""
        <!doctype html>
        <html>
        <head>
        <meta charset="{charset}">
        </head>
        <body>
        <p>{content}</p>
        </body>
        </html>
    """.strip()


@pytest.mark.asyncio
async def test_encode_from_headers():
    """Charset declared in headers takes precedence"""
    text = body_str(charset="gb2312")
    body, encoding = _encode_body(
        headers=Headers({"content-type": "text/html; charset=ISO-8859-1"}),
        text=text,
    )
    assert encoding == "cp1252"
    assert body == text.encode(encoding)


@pytest.mark.asyncio
async def test_encode_from_body():
    """No charset declared in headers, use the one declared in the body"""
    text = body_str(charset="gb2312")
    body, encoding = _encode_body(headers=Headers({}), text=text)
    assert encoding == "gb18030"
    assert body == text.encode(encoding)


@pytest.mark.asyncio
async def test_encode_fallback_utf8():
    """No charset declared, use utf-8 as fallback"""
    text = "<html>áéíóú</html>"
    body, encoding = _encode_body(headers=Headers(), text=text)
    assert encoding == "utf-8"
    assert body == text.encode(encoding)


@pytest.mark.asyncio
async def test_encode_mismatch():
    """Charset declared in headers and body do not match, and the headers
    one fails to encode: use the one in the body (first one that works)
    """
    text = body_str(charset="gb2312", content="空手道")
    body, encoding = _encode_body(
        headers=Headers({"content-type": "text/html; charset=ISO-8859-1"}),
        text=text,
    )
    assert encoding == "gb18030"
    assert body == text.encode(encoding)



import logging
import platform
import subprocess
from tempfile import NamedTemporaryFile

import pytest
from scrapy import Spider, Request
from scrapy.http.response.html import HtmlResponse

from scrapy_playwright.page import PageMethod

from tests import make_handler, assert_correct_response
from tests.mockserver import StaticMockServer


def get_mimetype(file):
    return subprocess.run(
        ["file", "--mime-type", "--brief", file.name],
        stdout=subprocess.PIPE,
        universal_newlines=True,
        check=False,
    ).stdout.strip()


@pytest.mark.asyncio
async def test_page_methods():
    screenshot = PageMethod("screenshot", "foo", 123, path="/tmp/file", type="png")
    assert screenshot.method == "screenshot"
    assert screenshot.args == ("foo", 123)
    assert screenshot.kwargs == {"path": "/tmp/file", "type": "png"}
    assert screenshot.result is None
    assert str(screenshot) == "<PageMethod for method 'screenshot'>"


class MixinPageMethodTestCase:
    @pytest.mark.asyncio
    async def test_page_non_page_method(self, caplog):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with StaticMockServer() as server:
                req = Request(
                    url=server.urljoin("/index.html"),
                    meta={
                        "playwright": True,
                        "playwright_page_methods": [
                            "not-a-page-method",
                            5,
                            None,
                        ],
                    },
                )
                resp = await handler._download_request(req, Spider("foo"))

        assert_correct_response(resp, req)
        for obj in req.meta["playwright_page_methods"]:
            assert (
                "scrapy-playwright",
                logging.WARNING,
                f"Ignoring {repr(obj)}: expected PageMethod, got {repr(type(obj))}",
            ) in caplog.record_tuples

    @pytest.mark.asyncio
    async def test_page_mixed_page_methods(self, caplog):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with StaticMockServer() as server:
                req = Request(
                    url=server.urljoin("/index.html"),
                    meta={
                        "playwright": True,
                        "playwright_page_methods": {
                            "does_not_exist": PageMethod("does_not_exist"),
                            "is_closed": PageMethod("is_closed"),  # not awaitable
                            "title": PageMethod("title"),  # awaitable
                        },
                    },
                )
                resp = await handler._download_request(req, Spider("foo"))

        assert_correct_response(resp, req)
        does_not_exist = req.meta["playwright_page_methods"]["does_not_exist"]
        assert (
            "scrapy-playwright",
            logging.WARNING,
            f"Ignoring {repr(does_not_exist)}: could not find method",
        ) in caplog.record_tuples
        assert not req.meta["playwright_page_methods"]["is_closed"].result
        assert req.meta["playwright_page_methods"]["title"].result == "Awesome site"

    @pytest.mark.asyncio
    async def test_page_method_navigation(self):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with StaticMockServer() as server:
                req = Request(
                    url=server.urljoin("/index.html"),
                    meta={
                        "playwright": True,
                        "playwright_page_methods": [PageMethod("click", "a.lorem_ipsum")],
                    },
                )
                resp = await handler._download_request(req, Spider("foo"))

            assert isinstance(resp, HtmlResponse)
            assert resp.request is req
            assert resp.url == server.urljoin("/lorem_ipsum.html")
            assert resp.status == 200
            assert "playwright" in resp.flags
            assert resp.css("title::text").get() == "Lorem Ipsum"
            text = resp.css("p::text").get()
            assert text == "Lorem ipsum dolor sit amet, consectetur adipiscing elit."

    @pytest.mark.asyncio
    async def test_page_method_infinite_scroll(self):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with StaticMockServer() as server:
                req = Request(
                    url=server.urljoin("/scroll.html"),
                    headers={"User-Agent": "scrapy-playwright"},
                    meta={
                        "playwright": True,
                        "playwright_page_methods": [
                            PageMethod("wait_for_selector", selector="div.quote"),
                            PageMethod(
                                "evaluate", "window.scrollBy(0, document.body.scrollHeight)"
                            ),
                            PageMethod("wait_for_selector", selector="div.quote:nth-child(11)"),
                            PageMethod(
                                "evaluate", "window.scrollBy(0, document.body.scrollHeight)"
                            ),
                            PageMethod("wait_for_selector", selector="div.quote:nth-child(21)"),
                        ],
                    },
                )
                resp = await handler._download_request(req, Spider("foo"))

            assert_correct_response(resp, req)
            assert len(resp.css("div.quote")) == 30

    @pytest.mark.asyncio
    async def test_page_method_screenshot(self):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with NamedTemporaryFile(mode="w+b") as png_file:
                with StaticMockServer() as server:
                    req = Request(
                        url=server.urljoin("/index.html"),
                        meta={
                            "playwright": True,
                            "playwright_page_methods": {
                                "png": PageMethod("screenshot", path=png_file.name, type="png"),
                            },
                        },
                    )
                    await handler._download_request(req, Spider("foo"))

                png_file.file.seek(0)
                assert png_file.file.read() == req.meta["playwright_page_methods"]["png"].result
                assert get_mimetype(png_file) == "image/png"

    @pytest.mark.asyncio
    async def test_page_method_pdf(self):
        if self.browser_type != "chromium":
            pytest.skip("PDF generation is supported only in Chromium")

        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with NamedTemporaryFile(mode="w+b") as pdf_file:
                with StaticMockServer() as server:
                    req = Request(
                        url=server.urljoin("/index.html"),
                        meta={
                            "playwright": True,
                            "playwright_page_methods": {
                                "pdf": PageMethod("pdf", path=pdf_file.name),
                            },
                        },
                    )
                    await handler._download_request(req, Spider("foo"))

                pdf_file.file.seek(0)
                assert pdf_file.file.read() == req.meta["playwright_page_methods"]["pdf"].result
                assert get_mimetype(pdf_file) == "application/pdf"


class TestPageMethodChromium(MixinPageMethodTestCase):
    browser_type = "chromium"


class TestPageMethodFirefox(MixinPageMethodTestCase):
    browser_type = "firefox"


@pytest.mark.skipif(platform.system() != "Darwin", reason="Test WebKit only on Darwin")
class TestPageMethodWebkit(MixinPageMethodTestCase):
    browser_type = "webkit"



import sys

import pytest
from scrapy import Spider
from scrapy.http import Request, Response
from scrapy.utils.test import get_crawler
from twisted.internet import defer
from twisted.trial.unittest import TestCase

from scrapy_playwright.handler import ScrapyPlaywrightDownloadHandler
from tests.mockserver import StaticMockServer


@pytest.mark.skipif(sys.version_info < (3, 8), reason="Times out with py37 and playwright>=1.10")
class MixedRequestsTestCase(TestCase):
    """
    This test case ensures the handler's 'download_request' method works as expected, and
    non-playwright requests are processed correctly. The rest of the tests directly call
    '_download_request', which is a coroutine ('download_request' returns a Deferred).
    This stopped working under py37 with playwright==1.10, but I couldn't find anything in the
    release notes to explain the change. Also, playwright doesn't currently define a __version__
    attribute, which would allow to enable the test for playwright<=1.9
    """

    @defer.inlineCallbacks
    def setUp(self):
        self.server = StaticMockServer()
        self.server.__enter__()
        self.handler = ScrapyPlaywrightDownloadHandler.from_crawler(get_crawler())
        yield self.handler._engine_started()

    @defer.inlineCallbacks
    def tearDown(self):
        self.server.__exit__(None, None, None)
        yield self.handler.close()

    def test_regular_request(self):
        def _test(response):
            self.assertIsInstance(response, Response)
            self.assertEqual(response.css("a::text").getall(), ["Lorem Ipsum", "Infinite Scroll"])
            self.assertEqual(response.url, request.url)
            self.assertEqual(response.status, 200)
            self.assertNotIn("playwright", response.flags)

        request = Request(self.server.urljoin("/index.html"))
        return self.handler.download_request(request, Spider("foo")).addCallback(_test)

    def test_playwright_request(self):
        def _test(response):
            self.assertIsInstance(response, Response)
            self.assertEqual(response.css("a::text").getall(), ["Lorem Ipsum", "Infinite Scroll"])
            self.assertEqual(response.url, request.url)
            self.assertEqual(response.status, 200)
            self.assertIn("playwright", response.flags)

        request = Request(self.server.urljoin("/index.html"), meta={"playwright": True})
        return self.handler.download_request(request, Spider("foo")).addCallback(_test)



import json
import logging
import platform
import sys
from ipaddress import ip_address
from unittest.mock import MagicMock, patch

import pytest
from playwright.async_api import (
    Dialog,
    Error as PlaywrightError,
    Page as PlaywrightPage,
    TimeoutError as PlaywrightTimeoutError,
)
from scrapy import Spider, Request, FormRequest
from scrapy.http import Response

from scrapy_playwright.handler import DEFAULT_CONTEXT_NAME
from scrapy_playwright.page import PageMethod

from tests import make_handler, assert_correct_response
from tests.mockserver import MockServer, StaticMockServer


class DialogSpider(Spider):
    """A spider with a method to handle the "dialog" page event"""

    name = "dialog"

    def parse(self, response: Response, **kwargs) -> None:
        return None

    async def handle_dialog(self, dialog: Dialog) -> None:
        self.dialog_message = dialog.message
        await dialog.dismiss()


class MixinTestCase:
    @pytest.mark.asyncio
    async def test_basic_response(self):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with StaticMockServer() as server:
                meta = {"playwright": True, "playwright_include_page": True}
                req = Request(server.urljoin("/index.html"), meta=meta)
                resp = await handler._download_request(req, Spider("foo"))

            assert_correct_response(resp, req)
            assert resp.css("a::text").getall() == ["Lorem Ipsum", "Infinite Scroll"]
            assert isinstance(resp.meta["playwright_page"], PlaywrightPage)
            assert resp.meta["playwright_page"].url == resp.url

            await resp.meta["playwright_page"].close()

    @pytest.mark.asyncio
    async def test_post_request(self):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with MockServer() as server:
                req = FormRequest(
                    server.urljoin("/"), meta={"playwright": True}, formdata={"foo": "bar"}
                )
                resp = await handler._download_request(req, Spider("foo"))

            assert_correct_response(resp, req)
            assert "Request body: foo=bar" in resp.text

    @pytest.mark.asyncio
    async def test_timeout_value(self):
        settings_dict = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
        }
        async with make_handler(settings_dict) as handler:
            assert handler.default_navigation_timeout is None

        settings_dict = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_DEFAULT_NAVIGATION_TIMEOUT": None,
        }
        async with make_handler(settings_dict) as handler:
            assert handler.default_navigation_timeout is None

        settings_dict = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_DEFAULT_NAVIGATION_TIMEOUT": 0,
        }
        async with make_handler(settings_dict) as handler:
            assert handler.default_navigation_timeout == 0

        settings_dict = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_DEFAULT_NAVIGATION_TIMEOUT": 123,
        }
        async with make_handler(settings_dict) as handler:
            assert handler.default_navigation_timeout == 123
        settings_dict = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_DEFAULT_NAVIGATION_TIMEOUT": 0.5,
        }
        async with make_handler(settings_dict) as handler:
            assert handler.default_navigation_timeout == 0.5

    @pytest.mark.asyncio
    async def test_timeout_error(self, caplog):
        settings_dict = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_DEFAULT_NAVIGATION_TIMEOUT": 100,
        }
        async with make_handler(settings_dict) as handler:
            with MockServer() as server:
                req = Request(server.urljoin("/delay/1"), meta={"playwright": True})
                with pytest.raises(PlaywrightTimeoutError) as excinfo:
                    await handler._download_request(req, Spider("foo"))
                assert (
                    "scrapy-playwright",
                    logging.WARNING,
                    f"Closing page due to failed request: {req}"
                    f" exc_type={type(excinfo.value)} exc_msg={str(excinfo.value)}",
                ) in caplog.record_tuples

    @pytest.mark.asyncio
    async def test_retry_page_content_still_navigating(self, caplog):
        if self.browser_type != "chromium":
            pytest.skip("Only Chromium seems to redirect meta tags within the same goto call")

        caplog.set_level(logging.DEBUG)
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with StaticMockServer() as server:
                req = Request(server.urljoin("/redirect.html"), meta={"playwright": True})
                resp = await handler._download_request(req, Spider("foo"))

            assert resp.request is req
            assert resp.url == server.urljoin("/index.html")  # redirected
            assert resp.status == 200
            assert "playwright" in resp.flags
            assert (
                "scrapy-playwright",
                logging.DEBUG,
                f"Retrying to get content from page '{req.url}', error: 'Unable to retrieve"
                " content because the page is navigating and changing the content.'",
            ) in caplog.record_tuples

    @pytest.mark.skipif(sys.version_info < (3, 8), reason="AsyncMock was added on Python 3.8")
    @patch("scrapy_playwright.handler.logger")
    @pytest.mark.asyncio
    async def test_route_continue_exception(self, logger):
        from unittest.mock import AsyncMock

        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            scrapy_request = Request(url="https://example.org", method="GET")
            spider = Spider("foo")
            req_handler = handler._make_request_handler(
                context_name=DEFAULT_CONTEXT_NAME,
                method=scrapy_request.method,
                url=scrapy_request.url,
                headers=scrapy_request.headers,
                body=None,
                encoding="utf-8",
                spider=spider,
            )
            route = MagicMock()
            playwright_request = AsyncMock()
            playwright_request.url = scrapy_request.url
            playwright_request.method = scrapy_request.method
            playwright_request.is_navigation_request = MagicMock(return_value=True)
            playwright_request.all_headers.return_value = {}

            # safe error, only warn
            ex = PlaywrightError("Target page, context or browser has been closed")
            route.continue_.side_effect = ex
            await req_handler(route, playwright_request)
            logger.warning.assert_called_with(
                "Failed processing Playwright request: <%s %s> exc_type=%s exc_msg=%s",
                playwright_request.method,
                playwright_request.url,
                type(ex),
                str(ex),
                extra={
                    "spider": spider,
                    "context_name": DEFAULT_CONTEXT_NAME,
                    "scrapy_request_url": scrapy_request.url,
                    "scrapy_request_method": scrapy_request.method,
                    "playwright_request_url": playwright_request.url,
                    "playwright_request_method": playwright_request.method,
                    "exception": ex,
                },
            )

            # unknown errors, re-raise
            route.continue_.side_effect = ZeroDivisionError("asdf")
            with pytest.raises(ZeroDivisionError):
                await req_handler(route, playwright_request)
            route.continue_.side_effect = PlaywrightError("qwerty")
            with pytest.raises(PlaywrightError):
                await req_handler(route, playwright_request)

    @pytest.mark.asyncio
    async def test_event_handler_dialog_callable(self):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with StaticMockServer() as server:
                spider = DialogSpider()
                req = Request(
                    url=server.urljoin("/index.html"),
                    meta={
                        "playwright": True,
                        "playwright_page_methods": [
                            # trigger an alert
                            PageMethod("evaluate", "alert('foobar');"),
                        ],
                        "playwright_page_event_handlers": {
                            "dialog": spider.handle_dialog,
                        },
                    },
                )
                await handler._download_request(req, spider)

            assert spider.dialog_message == "foobar"

    @pytest.mark.asyncio
    async def test_event_handler_dialog_str(self):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with StaticMockServer() as server:
                spider = DialogSpider()
                req = Request(
                    url=server.urljoin("/index.html"),
                    meta={
                        "playwright": True,
                        "playwright_page_methods": [
                            # trigger an alert
                            PageMethod("evaluate", "alert('foobar');"),
                        ],
                        "playwright_page_event_handlers": {
                            "dialog": "handle_dialog",
                        },
                    },
                )
                await handler._download_request(req, spider)

            assert spider.dialog_message == "foobar"

    @pytest.mark.asyncio
    async def test_event_handler_dialog_missing(self, caplog):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with StaticMockServer() as server:
                spider = DialogSpider()
                req = Request(
                    url=server.urljoin("/index.html"),
                    meta={
                        "playwright": True,
                        "playwright_page_event_handlers": {
                            "dialog": "missing_method",
                        },
                    },
                )
                await handler._download_request(req, spider)

        assert (
            "scrapy-playwright",
            logging.WARNING,
            "Spider 'dialog' does not have a 'missing_method' attribute,"
            " ignoring handler for event 'dialog'",
        ) in caplog.record_tuples
        assert getattr(spider, "dialog_message", None) is None

    @pytest.mark.asyncio
    async def test_response_attributes(self):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with MockServer() as server:
                req = Request(
                    url=server.urljoin(),
                    meta={"playwright": True},
                )
                response = await handler._download_request(req, Spider("spider_name"))

        assert response.ip_address == ip_address(server.address)

    @pytest.mark.asyncio
    async def test_page_goto_kwargs_referer(self):
        if self.browser_type != "chromium":
            pytest.skip("referer as goto kwarg seems to work only with chromium :shrug:")
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with MockServer() as server:
                fake_referer = server.urljoin("/fake/referer")
                req = Request(
                    url=server.urljoin("/headers"),
                    meta={
                        "playwright": True,
                        "playwright_page_goto_kwargs": {"referer": fake_referer},
                    },
                )
                response = await handler._download_request(req, Spider("spider_name"))

        headers = json.loads(response.css("pre::text").get())
        assert headers["Referer"] == fake_referer

    @pytest.mark.asyncio
    async def test_navigation_returns_none(self, caplog):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with MockServer():
                req = Request(url="about:blank", meta={"playwright": True})
                response = await handler._download_request(req, Spider("spider_name"))

        assert (
            "scrapy-playwright",
            logging.WARNING,
            f"Navigating to {req!r} returned None, the response"
            " will have empty headers and status 200",
        ) in caplog.record_tuples
        assert not response.headers
        assert response.status == 200

    @pytest.mark.asyncio
    async def test_abort_requests(self):
        async def should_abort_request_async(request):
            return request.resource_type == "image"

        def should_abort_request_sync(request):
            return request.resource_type == "image"

        for predicate in (
            lambda request: request.resource_type == "image",
            should_abort_request_async,
            should_abort_request_sync,
        ):
            settings_dict = {
                "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
                "PLAYWRIGHT_ABORT_REQUEST": predicate,
            }
            async with make_handler(settings_dict) as handler:
                with StaticMockServer() as server:
                    req = Request(
                        url=server.urljoin("/gallery.html"),
                        meta={"playwright": True},
                    )
                    await handler._download_request(req, Spider("foo"))

                    req_prefix = "playwright/request_count"
                    resp_prefix = "playwright/response_count"
                    assert handler.stats.get_value(f"{req_prefix}/resource_type/document") == 1
                    assert handler.stats.get_value(f"{req_prefix}/resource_type/image") == 3
                    assert handler.stats.get_value(f"{resp_prefix}/resource_type/document") == 1
                    assert handler.stats.get_value(f"{resp_prefix}/resource_type/image") is None
                    assert handler.stats.get_value(f"{req_prefix}/aborted") == 3

    @pytest.mark.asyncio
    async def test_page_initialization_ok(self):
        async def init_page(page, _request):
            await page.set_extra_http_headers({"Extra-Header": "Qwerty"})

        settings_dict = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_PROCESS_REQUEST_HEADERS": None,
        }
        async with make_handler(settings_dict) as handler:
            with MockServer() as server:
                req = Request(
                    url=server.urljoin("/headers"),
                    meta={"playwright": True, "playwright_page_init_callback": init_page},
                )
                response = await handler._download_request(req, Spider("spider_name"))
        assert response.status == 200
        headers = json.loads(response.css("pre::text").get())
        headers = {key.lower(): value for key, value in headers.items()}
        assert headers["extra-header"] == "Qwerty"

    @pytest.mark.asyncio
    async def test_page_initialization_fail(self, caplog):
        async def init_page(page, _request, _missing):
            await page.set_extra_http_headers({"Extra-Header": "Qwerty"})

        settings_dict = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_PROCESS_REQUEST_HEADERS": None,
        }
        async with make_handler(settings_dict) as handler:
            with MockServer() as server:
                req = Request(
                    url=server.urljoin("/headers"),
                    meta={"playwright": True, "playwright_page_init_callback": init_page},
                )
                response = await handler._download_request(req, Spider("spider_name"))
        assert response.status == 200
        headers = json.loads(response.css("pre::text").get())
        headers = {key.lower(): value for key, value in headers.items()}
        assert "extra-header" not in headers

        log_entry = caplog.record_tuples[0]
        assert log_entry[0] == "scrapy-playwright"
        assert log_entry[1] == logging.WARNING
        assert f"[Context=default] Page init callback exception for {req!r}" in log_entry[2]
        assert "init_page() missing 1 required positional argument: '_missing'" in log_entry[2]

    @pytest.mark.asyncio
    async def test_redirect(self):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with MockServer() as server:
                req = Request(
                    url=server.urljoin("/redirect2"),
                    meta={"playwright": True},
                )
                response = await handler._download_request(req, Spider("spider_name"))

        assert response.url == server.urljoin("/headers")
        assert response.meta["redirect_times"] == 2
        assert response.meta["redirect_reasons"] == [302, 301]
        assert response.meta["redirect_urls"] == [
            server.urljoin("/redirect2"),
            server.urljoin("/redirect"),
        ]

    @pytest.mark.asyncio
    async def test_logging_record_spider(self, caplog):
        """Make sure at least one log record has the spider as an attribute
        (records sent before opening the spider will not have it).
        """
        caplog.set_level(logging.DEBUG)
        spider = Spider("spider_name")
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            with MockServer() as server:
                req = Request(url=server.urljoin("/index.html"), meta={"playwright": True})
                await handler._download_request(req, spider)

        assert any(getattr(rec, "spider", None) is spider for rec in caplog.records)


class TestCaseChromium(MixinTestCase):
    browser_type = "chromium"


class TestCaseFirefox(MixinTestCase):
    browser_type = "firefox"


@pytest.mark.skipif(platform.system() != "Darwin", reason="Test WebKit only on Darwin")
class TestCaseWebkit(MixinTestCase):
    browser_type = "webkit"



import asyncio
import platform
import tempfile
from pathlib import Path
from uuid import uuid4

import pytest
from playwright.async_api import Browser, TimeoutError as PlaywrightTimeoutError
from scrapy import Spider, Request
from scrapy_playwright.page import PageMethod

from tests import make_handler
from tests.mockserver import StaticMockServer


class MixinTestCaseMultipleContexts:
    @pytest.mark.asyncio
    async def test_contexts_max_settings(self):
        settings = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_MAX_PAGES_PER_CONTEXT": 1234,
        }
        async with make_handler(settings) as handler:
            assert handler.max_pages_per_context == 1234

        settings = {"PLAYWRIGHT_BROWSER_TYPE": self.browser_type, "CONCURRENT_REQUESTS": 9876}
        async with make_handler(settings) as handler:
            assert handler.max_pages_per_context == 9876

        settings = {"PLAYWRIGHT_BROWSER_TYPE": self.browser_type, "PLAYWRIGHT_MAX_CONTEXTS": None}
        async with make_handler(settings) as handler:
            assert not hasattr(handler, "context_semaphore")

        settings = {"PLAYWRIGHT_BROWSER_TYPE": self.browser_type, "PLAYWRIGHT_MAX_CONTEXTS": 1234}
        async with make_handler(settings) as handler:
            assert handler.context_semaphore._value == 1234

    @pytest.mark.asyncio
    async def test_context_kwargs(self):
        settings_dict = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_CONTEXTS": {
                "default": {"java_script_enabled": False},
            },
        }
        async with make_handler(settings_dict) as handler:
            with StaticMockServer() as server:
                req = Request(
                    url=server.urljoin("/scroll.html"),
                    meta={
                        "playwright": True,
                        "playwright_page_methods": [
                            # cause a timeout by waiting on an element that is rendered with js
                            PageMethod("wait_for_selector", selector="div.quote", timeout=1000),
                        ],
                    },
                )
                with pytest.raises(PlaywrightTimeoutError):
                    await handler._download_request(req, Spider("foo"))

    @pytest.mark.asyncio
    async def test_contexts_max_pages(self):
        settings = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_MAX_PAGES_PER_CONTEXT": 2,
            "PLAYWRIGHT_CONTEXTS": {
                "a": {"java_script_enabled": True},
                "b": {"java_script_enabled": True},
            },
        }
        async with make_handler(settings) as handler:
            with StaticMockServer() as server:
                requests = [
                    handler._download_request(
                        Request(
                            server.urljoin(f"/index.html?a={i}"),
                            meta={"playwright": True, "playwright_context": "a"},
                        ),
                        Spider("foo"),
                    )
                    for i in range(20)
                ] + [
                    handler._download_request(
                        Request(
                            server.urljoin(f"/index.html?b={i}"),
                            meta={"playwright": True, "playwright_context": "b"},
                        ),
                        Spider("foo"),
                    )
                    for i in range(20)
                ]
                await asyncio.gather(*requests)

            assert handler.stats.get_value("playwright/page_count/max_concurrent") == 4

    @pytest.mark.asyncio
    async def test_max_contexts(self):
        def cb_close_context(task):
            response = task.result()
            asyncio.create_task(response.meta["playwright_page"].context.close())

        settings = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_MAX_CONTEXTS": 4,
        }
        async with make_handler(settings) as handler:
            with StaticMockServer() as server:
                tasks = []
                for i in range(20):
                    request = Request(
                        url=server.urljoin(f"/index.html?a={i}"),
                        callback=cb_close_context,
                        meta={
                            "playwright": True,
                            "playwright_include_page": True,
                            "playwright_context": f"ctx-{i}",
                        },
                    )
                    coro = handler._download_request(
                        request=request,
                        spider=Spider("foo"),
                    )
                    # callbacks are not invoked at the download handler, call them explicitly
                    task = asyncio.create_task(coro)
                    task.add_done_callback(request.callback)
                    tasks.append(task)
                await asyncio.gather(*tasks)

            assert handler.stats.get_value("playwright/context_count/max_concurrent") == 4

    @pytest.mark.asyncio
    async def test_contexts_startup(self):
        settings = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_CONTEXTS": {
                "first": {
                    "storage_state": {
                        "cookies": [
                            {
                                "url": "https://example.org",
                                "name": "foo",
                                "value": "bar",
                            },
                        ],
                    },
                },
            },
        }
        async with make_handler(settings) as handler:
            assert len(handler.context_wrappers) == 1

            with StaticMockServer() as server:
                meta = {
                    "playwright": True,
                    "playwright_include_page": True,
                    "playwright_context": "first",
                }
                req = Request(server.urljoin("/index.html"), meta=meta)
                resp = await handler._download_request(req, Spider("foo"))

            page = resp.meta["playwright_page"]
            storage_state = await page.context.storage_state()
            await page.close()
            await page.context.close()
            cookie = storage_state["cookies"][0]
            assert cookie["name"] == "foo"
            assert cookie["value"] == "bar"
            assert cookie["domain"] == "example.org"

    @pytest.mark.asyncio
    async def test_persistent_context(self):
        temp_dir = f"{tempfile.gettempdir()}/{uuid4()}"
        settings = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_DEFAULT_NAVIGATION_TIMEOUT": 3000,
            "PLAYWRIGHT_CONTEXTS": {
                "persistent": {
                    "user_data_dir": temp_dir,
                },
            },
        }
        assert not Path(temp_dir).exists()
        async with make_handler(settings) as handler:
            assert Path(temp_dir).is_dir()
            assert len(handler.context_wrappers) == 1
            assert handler.context_wrappers["persistent"].persistent
            assert not hasattr(handler, "browser")

    @pytest.mark.asyncio
    async def test_mixed_persistent_contexts(self):
        temp_dir = f"{tempfile.gettempdir()}/{uuid4()}"
        settings = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_DEFAULT_NAVIGATION_TIMEOUT": 3000,
            "PLAYWRIGHT_CONTEXTS": {
                "persistent": {
                    "user_data_dir": temp_dir,
                },
                "non-persistent": {
                    "java_script_enabled": False,
                },
            },
        }
        assert not Path(temp_dir).exists()
        async with make_handler(settings) as handler:
            assert Path(temp_dir).is_dir()
            assert len(handler.context_wrappers) == 2
            assert handler.context_wrappers["persistent"].persistent
            assert not handler.context_wrappers["non-persistent"].persistent
            assert isinstance(handler.browser, Browser)

    @pytest.mark.asyncio
    async def test_contexts_dynamic(self):
        async with make_handler({"PLAYWRIGHT_BROWSER_TYPE": self.browser_type}) as handler:
            assert len(handler.context_wrappers) == 0

            with StaticMockServer() as server:
                meta = {
                    "playwright": True,
                    "playwright_include_page": True,
                    "playwright_context": "new",
                    "playwright_context_kwargs": {
                        "storage_state": {
                            "cookies": [
                                {
                                    "url": "https://example.org",
                                    "name": "asdf",
                                    "value": "qwerty",
                                },
                            ],
                        },
                    },
                }
                req = Request(server.urljoin("/index.html"), meta=meta)
                resp = await handler._download_request(req, Spider("foo"))

            assert len(handler.context_wrappers) == 1

            page = resp.meta["playwright_page"]
            storage_state = await page.context.storage_state()
            await page.close()
            cookie = storage_state["cookies"][0]
            assert cookie["name"] == "asdf"
            assert cookie["value"] == "qwerty"
            assert cookie["domain"] == "example.org"


class TestCaseMultipleContextsChromium(MixinTestCaseMultipleContexts):
    browser_type = "chromium"


class TestCaseMultipleContextsFirefox(MixinTestCaseMultipleContexts):
    browser_type = "firefox"


@pytest.mark.skipif(platform.system() != "Darwin", reason="Test WebKit only on Darwin")
class TestCaseMultipleContextsWebkit(MixinTestCaseMultipleContexts):
    browser_type = "webkit"



import json
import platform

import pytest
from scrapy import Spider, Request

from tests import make_handler
from tests.mockserver import MockServer


class MixinProcessHeadersTestCase:
    @pytest.mark.asyncio
    async def test_user_agent(self):
        settings_dict = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_CONTEXTS": {"default": {"user_agent": self.browser_type}},
            "USER_AGENT": None,
        }
        async with make_handler(settings_dict) as handler:
            with MockServer() as server:
                # if Scrapy's user agent is None, use the one from the Browser
                req = Request(
                    url=server.urljoin("/headers"),
                    meta={"playwright": True},
                )
                resp = await handler._download_request(req, Spider("foo"))
                headers = json.loads(resp.css("pre::text").get())
                headers = {key.lower(): value for key, value in headers.items()}
                assert headers["user-agent"] == self.browser_type

                # if Scrapy's user agent is set to some value, use it
                req = Request(
                    url=server.urljoin("/headers"),
                    meta={"playwright": True},
                    headers={"User-Agent": "foobar"},
                )
                resp = await handler._download_request(req, Spider("foo"))
                headers = json.loads(resp.css("pre::text").get())
                headers = {key.lower(): value for key, value in headers.items()}
                assert headers["user-agent"] == "foobar"

    @pytest.mark.asyncio
    async def test_playwright_headers(self):
        """Ignore Scrapy headers"""
        settings_dict = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_CONTEXTS": {"default": {"user_agent": self.browser_type}},
            "PLAYWRIGHT_PROCESS_REQUEST_HEADERS": None,
            "PLAYWRIGHT_DEFAULT_NAVIGATION_TIMEOUT": 2000,
        }
        async with make_handler(settings_dict) as handler:
            with MockServer() as server:
                req = Request(
                    url=server.urljoin("/headers"),
                    meta={"playwright": True},
                    headers={"User-Agent": "foobar", "Asdf": "qwerty"},
                )
                resp = await handler._download_request(req, Spider("foo"))
                headers = json.loads(resp.css("pre::text").get())
                headers = {key.lower(): value for key, value in headers.items()}
                assert headers["user-agent"] == self.browser_type
                assert req.headers["user-agent"].decode("utf-8") == self.browser_type
                assert "asdf" not in headers
                assert "asdf" not in req.headers
                assert b"asdf" not in req.headers

    @pytest.mark.asyncio
    async def test_use_custom_headers(self):
        """Custom header processing function"""

        async def important_headers(*_args, **_kwargs) -> dict:
            return {"foo": "bar"}

        settings_dict = {
            "PLAYWRIGHT_BROWSER_TYPE": self.browser_type,
            "PLAYWRIGHT_CONTEXTS": {"default": {"user_agent": self.browser_type}},
            "PLAYWRIGHT_PROCESS_REQUEST_HEADERS": important_headers,
        }
        async with make_handler(settings_dict) as handler:
            with MockServer() as server:
                req = Request(
                    url=server.urljoin("/headers"),
                    meta={"playwright": True},
                    headers={"User-Agent": "foobar", "Asdf": "qwerty"},
                )
                resp = await handler._download_request(req, Spider("foo"))
                headers = json.loads(resp.css("pre::text").get())
                headers = {key.lower(): value for key, value in headers.items()}
                assert headers["foo"] == "bar"
                assert headers.get("user-agent") not in (self.browser_type, "foobar")
                assert "asdf" not in headers


class TestProcessHeadersChromium(MixinProcessHeadersTestCase):
    browser_type = "chromium"


class TestProcessHeadersFirefox(MixinProcessHeadersTestCase):
    browser_type = "firefox"


@pytest.mark.skipif(platform.system() != "Darwin", reason="Test WebKit only on Darwin")
class TestProcessHeadersWebkit(MixinProcessHeadersTestCase):
    browser_type = "webkit"
